/*
 * mesg.c
 *
 * A library for processing quake network messages
 *
 * Kekoa Proudfoot 11/29/96
 */

#include <stdlib.h>
#include <stdio.h>
#include "mesg.h"
#include "endian.h"
#include "debug.h"
#include "hexdump.h"

/*****************************************************************************
 * Macro definitions
 *****************************************************************************/

/* Get macros */

#define GET_CHAR(b)        (b += 1, b[-1])
#define GET_SHORT(b)       (b += 2, read_little_short(&b[-2]))
#define GET_LONG(b)        (b += 4, read_little_long (&b[-4]))
#define GET_FLOAT(b)       (b += 4, read_little_float(&b[-4]))

#define GET_ANGLE(b)       (b += 1, b[-1] * (360.0f / 256.0f))
#define GET_COORD(b)       (b += 2, (short)read_little_short(&b[-2]) * 0.125f)
#define GET_CVELOCITY(b)   (b += 1, b[-1] * (256.0f / 8.0f))

#define GET_VOLUME(b)      (b += 1, b[-1] * (1.0f / 255.0f))

#define GET_STRING(b)      b; while (*b++)

/* Put macros */

#define ROUNDP(v)          (((v) - (int)(v) < 0.5) ? (int)(v) : (int)(v) + 1)
#define ROUNDN(v)          (((int)(v) - (v) < 0.5) ? (int)(v) : (int)(v) - 1)
#define ROUND(v)           ((v > 0.0) ? ROUNDP(v) : ROUNDN(v))

#define PUT_CHAR(b,v)      (b[0] = v,                               b += 1)
#define PUT_SHORT(b,v)     (write_little_short(b, v),               b += 2)
#define PUT_LONG(b,v)      (write_little_long (b, v),               b += 4)
#define PUT_FLOAT(b,v)     (write_little_float(b, v),               b += 4)

#define PUT_ANGLE(b,v)     (b[0] = ROUND(v * (256.0f / 360.0f)),    b += 1)
#define PUT_COORD(b,v)     (write_little_short(b, (short)ROUND(v * 8.0f)), b += 2)
#define PUT_CVELOCITY(b,v) (b[0] = ROUND(v * (8.0f / 256.0f)),      b += 1)

#define PUT_VOLUME(b,v)    (b[0] = ROUND(v * 255.0f),               b += 1)

#define PUT_STRING(b,v)    { char *c = v; do { *b = *c++; } while (*b++); }

/* Skip macros */

#define SKIP_CHAR(b)       b += 1
#define SKIP_SHORT(b)      b += 2
#define SKIP_LONG(b)       b += 4
#define SKIP_FLOAT(b)      b += 4

#define SKIP_ANGLE(b)      b += 1
#define SKIP_COORD(b)      b += 2
#define SKIP_CVELOCITY(b)  b += 1

#define SKIP_VOLUME(b)     b += 1

#define SKIP_STRING(b)     while (*b++)

/*****************************************************************************
 * Interface function definitions
 *****************************************************************************/

int
cl_to_sv_mesg_get(char *buf, cl_to_sv_mesg_t *mesg)
{
    char *start = buf;

    switch (*buf) {
    case MESG_CLIENT_CONNECT:
	mesg->client_connect.type         = GET_CHAR(buf);
	break;
    case MESG_CLIENT_DISCONNECT:
	mesg->client_disconnect.type      = GET_CHAR(buf);
	break;
    case MESG_SERVER_DATA:
	mesg->server_data.type            = GET_CHAR(buf);
	mesg->server_data.time            = GET_FLOAT(buf);
	mesg->server_data.orientation.p   = GET_ANGLE(buf);
	mesg->server_data.orientation.y   = GET_ANGLE(buf);
	mesg->server_data.orientation.r   = GET_ANGLE(buf);
	mesg->server_data.velocity.x      = (short)GET_SHORT(buf);
	mesg->server_data.velocity.y      = (short)GET_SHORT(buf);
	mesg->server_data.velocity.z      = (short)GET_SHORT(buf);
	mesg->server_data.flags           = GET_CHAR(buf);
	mesg->server_data.impulse         = GET_CHAR(buf);
	break;
    case MESG_EXECUTE_COMMAND:
	mesg->execute_command.type        = GET_CHAR(buf);
	mesg->execute_command.text        = GET_STRING(buf);
	break;
    default:
	mesg->bad.type                    = MESG_BAD;
	break;
    }

    return buf - start;
}

int
cl_to_sv_mesg_put(char *buf, cl_to_sv_mesg_t *mesg)
{
    char *start = buf;

    switch (mesg->type) {
    case MESG_CLIENT_CONNECT:
	PUT_CHAR  (buf, mesg->client_connect.type);
	break;
    case MESG_CLIENT_DISCONNECT:
	PUT_CHAR  (buf, mesg->client_disconnect.type);
	break;
    case MESG_SERVER_DATA:
	PUT_CHAR  (buf, mesg->server_data.type);
	PUT_FLOAT (buf, mesg->server_data.time);
	PUT_ANGLE (buf, mesg->server_data.orientation.p);
	PUT_ANGLE (buf, mesg->server_data.orientation.y);
	PUT_ANGLE (buf, mesg->server_data.orientation.r);
	PUT_SHORT (buf, (short)mesg->server_data.velocity.x);
	PUT_SHORT (buf, (short)mesg->server_data.velocity.y);
	PUT_SHORT (buf, (short)mesg->server_data.velocity.z);
	PUT_CHAR  (buf, mesg->server_data.flags);
	PUT_CHAR  (buf, mesg->server_data.impulse);
	break;
    case MESG_EXECUTE_COMMAND:
	PUT_CHAR  (buf, mesg->execute_command.type);
	PUT_STRING(buf, mesg->execute_command.text);
	break;
    default:
	PUT_CHAR  (buf, MESG_BAD);
	break;
    }

    return buf - start;
}

char *
cl_to_sv_mesg_name(cl_to_sv_mesg_t *mesg)
{
    switch (mesg->type) {
    default:
    case MESG_BAD:
	return "MESG_BAD";
    case MESG_CLIENT_CONNECT:
	return "MESG_CLIENT_CONNECT";
    case MESG_CLIENT_DISCONNECT:
	return "MESG_CLIENT_DISCONNECT";
    case MESG_SERVER_DATA:
	return "MESG_SERVER_DATA";
    case MESG_EXECUTE_COMMAND:
	return "MESG_EXECUTE_COMMAND";
    }
}

void
cl_to_sv_mesg_debug(cl_to_sv_mesg_t *mesg, FILE *file)
{
    char buf[16384];
    int  len;

    switch (mesg->type) {
    default:
	len = cl_to_sv_mesg_put(buf, mesg);
	fprintf(file, "[client message: %s, %d bytes]\n",
		cl_to_sv_mesg_name(mesg), len);
	hex_dump(buf, len, "C", file);
	break;
    }
}

char
cl_to_sv_mesg_peek(char *buf)
{
    switch (*buf) {
    default:
    case MESG_BAD:
	return MESG_BAD;
    case MESG_CLIENT_CONNECT:
    case MESG_CLIENT_DISCONNECT:
    case MESG_SERVER_DATA:
    case MESG_EXECUTE_COMMAND:
	return *buf;
    }
}

int
cl_to_sv_mesg_skip(char *buf)
{
    char *start = buf;

    switch (*buf) {
    case MESG_CLIENT_CONNECT:
	SKIP_CHAR  (buf);
	break;
    case MESG_CLIENT_DISCONNECT:
	SKIP_CHAR  (buf);
	break;
    case MESG_SERVER_DATA:
	SKIP_CHAR  (buf);
	SKIP_FLOAT (buf);
	SKIP_ANGLE (buf);
	SKIP_ANGLE (buf);
	SKIP_ANGLE (buf);
	SKIP_SHORT (buf);
	SKIP_SHORT (buf);
	SKIP_SHORT (buf);
	SKIP_CHAR  (buf);
	SKIP_CHAR  (buf);
	break;
    case MESG_EXECUTE_COMMAND:
	SKIP_CHAR  (buf);
	SKIP_STRING(buf);
	break;
    default:
	break;
    }

    return buf - start;
}

int
sv_to_cl_mesg_get(char *buf, sv_to_cl_mesg_t *mesg)
{
    short temp;
    short mask;
    char *start = buf;

    switch (*buf) {
    case MESG_NOP:
	mesg->nop.type                            = GET_CHAR(buf);
	break;
    case MESG_DISCONNECT:
	mesg->disconnect.type                     = GET_CHAR(buf);
	break;
    case MESG_UPDATE_STAT:
	mesg->update_stat.type                    = GET_CHAR(buf);
	mesg->update_stat.index                   = GET_CHAR(buf);
	mesg->update_stat.value                   = GET_LONG(buf);
	break;
    case MESG_VERSION:
	mesg->version.type                        = GET_CHAR(buf);
	mesg->version.version                     = GET_LONG(buf);
	break;
    case MESG_SET_VIEW:
	mesg->set_view.type                       = GET_CHAR(buf);
	mesg->set_view.entity                     = GET_SHORT(buf);
	break;
    case MESG_SOUND:
	mesg->sound.type                          = GET_CHAR(buf);
	mesg->sound.mask = mask                   = GET_CHAR(buf);
	mesg->sound.volume                        = (mask & 0x0001) ?
	                                            GET_VOLUME(buf) : 1.0;
	mesg->sound.attenuation                   = (mask & 0x0002) ?
	                                            GET_CHAR(buf) >> 6 : 1;
	temp                                      = GET_SHORT(buf);
	mesg->sound.channel                       = temp & 0x07;
	mesg->sound.entity                        = (temp >> 3) & 0x1fff;
	mesg->sound.sound                         = GET_CHAR(buf);
	mesg->sound.position.x                    = GET_COORD(buf);
	mesg->sound.position.y                    = GET_COORD(buf);
	mesg->sound.position.z                    = GET_COORD(buf);
	break ;
    case MESG_TIME:
	mesg->time.type                           = GET_CHAR(buf);
	mesg->time.time                           = GET_FLOAT(buf);
	break;
    case MESG_PRINT:
	mesg->print.type                          = GET_CHAR(buf);
	mesg->print.text                          = GET_STRING(buf);
	break;
    case MESG_STUFF_TEXT:
	mesg->stuff_text.type                     = GET_CHAR(buf);
	mesg->stuff_text.text                     = GET_STRING(buf);
	break;
    case MESG_SET_ANGLE:
	mesg->set_angle.type                      = GET_CHAR(buf);
	mesg->set_angle.orientation.p             = GET_ANGLE(buf);
	mesg->set_angle.orientation.y             = GET_ANGLE(buf);
	mesg->set_angle.orientation.r             = GET_ANGLE(buf);
	break;
    case MESG_SERVER_INFO:
	mesg->server_info.type                    = GET_CHAR(buf);
	mesg->server_info.version                 = GET_LONG(buf);
	mesg->server_info.max_clients             = GET_CHAR(buf);
	mesg->server_info.multi_player            = GET_CHAR(buf);
	mesg->server_info.map_name                = GET_STRING(buf);
	mesg->server_info.num_models              = 0;
	do {
	    mesg->server_info.num_models++;
	    if (mesg->server_info.num_models < MESG_NUM_MODEL_NAMES) {
		mesg->server_info.model_names[mesg->server_info.num_models]
		                                  = GET_STRING(buf);
	    }
	    else {
		mesg->server_info.num_models--;
		SKIP_STRING(buf);
		debug_print(DEBUG_WARNING,
			    "sv_to_cl_mesg_get: model name overflow");
	    }
	} while (*buf);
	buf++;
	mesg->server_info.num_sounds              = 0;
	do {
	    mesg->server_info.num_sounds++;
	    if (mesg->server_info.num_sounds < MESG_NUM_SOUND_NAMES) {
		mesg->server_info.sound_names[mesg->server_info.num_sounds]
		                                  = GET_STRING(buf);
	    }
	    else {
		mesg->server_info.num_sounds--;
		SKIP_STRING(buf);
		debug_print(DEBUG_WARNING,
			    "sv_to_cl_mesg_get: sound name overflow");
	    }
	} while (*buf);
	buf++;
	break;
    case MESG_LIGHT_STYLE:
	mesg->light_style.type                    = GET_CHAR(buf);
	mesg->light_style.style                   = GET_CHAR(buf);
	mesg->light_style.text                    = GET_STRING(buf);
	break;
    case MESG_UPDATE_NAME:
	mesg->update_name.type                    = GET_CHAR(buf);
	mesg->update_name.player                  = GET_CHAR(buf);
	mesg->update_name.name                    = GET_STRING(buf);
	break;
    case MESG_UPDATE_FRAGS:
	mesg->update_frags.type                   = GET_CHAR(buf);
	mesg->update_frags.player                 = GET_CHAR(buf);
	mesg->update_frags.frags                  = GET_SHORT(buf);
	break;
    case MESG_CLIENT_DATA:
        /* Apparently, newer versions of quake (post 1.06), make the
         * 'items' field compulsory. Since I'm not interpreting that anyway,
         * I'll just get the long everytime.
         */
	mesg->client_data.type                    = GET_CHAR(buf);
	mesg->client_data.mask = mask             = GET_SHORT(buf);
	mesg->client_data.view_z_offset           = (mask & 0x0001) ?
	                                            GET_CHAR(buf) : 22;
	mesg->client_data.view_pitch_offset       = (mask & 0x0002) ?
	                                            GET_CHAR(buf) : 0;
	mesg->client_data.orientation.p           = (mask & 0x0004) ?
	                                            GET_ANGLE(buf) : 0;
	mesg->client_data.velocity.x              = (mask & 0x0020) ?
	                                            GET_CVELOCITY(buf) : 0;
	mesg->client_data.orientation.y           = (mask & 0x0008) ?
	                                            GET_ANGLE(buf) : 0;
	mesg->client_data.velocity.y              = (mask & 0x0040) ?
	                                            GET_CVELOCITY(buf) : 0;
	mesg->client_data.orientation.r           = (mask & 0x0010) ?
	                                            GET_ANGLE(buf) : 0.0f;
	mesg->client_data.velocity.z              = (mask & 0x0080) ?
	                                            GET_CVELOCITY(buf) : 0.0f;
/* 	mesg->client_data.items                   = (mask & 0x0200) ? */
/* 	                                            GET_LONG(buf) : 0x1001; */
	mesg->client_data.items                   = GET_LONG(buf);
	mesg->client_data.bits                    = (mask & 0x0c00) >> 10;
	mesg->client_data.weapon_frame            = (mask & 0x1000) ?
	                                            GET_CHAR(buf) : 0;
	mesg->client_data.armor                   = (mask & 0x2000) ?
	                                            GET_CHAR(buf) : 0;
	mesg->client_data.weapon_model            = (mask & 0x4000) ?
	                                            GET_CHAR(buf) : 0;
	mesg->client_data.health                  = GET_SHORT(buf);
	mesg->client_data.current_ammo            = GET_CHAR(buf);
	mesg->client_data.ammo_shells             = GET_CHAR(buf);
	mesg->client_data.ammo_nails              = GET_CHAR(buf);
	mesg->client_data.ammo_rockets            = GET_CHAR(buf);
	mesg->client_data.ammo_cells              = GET_CHAR(buf);
	mesg->client_data.current_weapon          = GET_CHAR(buf);
	break;
    case MESG_STOP_SOUND:
	mesg->stop_sound.type                     = GET_CHAR(buf);
	mesg->stop_sound.sound                    = GET_SHORT(buf);
	break;
    case MESG_UPDATE_COLORS:
	mesg->update_colors.type                  = GET_CHAR(buf);
	mesg->update_colors.player                = GET_CHAR(buf);
	temp                                      = GET_CHAR(buf);
	mesg->update_colors.shirt_color           = (temp >> 4) & 0x0f;
	mesg->update_colors.pants_color           = (temp >> 0) & 0x0f; 
	break;
    case MESG_PARTICLE:
	mesg->particle.type                       = GET_CHAR(buf);
	mesg->particle.position.x                 = GET_COORD(buf);
	mesg->particle.position.y                 = GET_COORD(buf);
	mesg->particle.position.z                 = GET_COORD(buf);
	mesg->particle.velocity.x                 = GET_CHAR(buf);
	mesg->particle.velocity.y                 = GET_CHAR(buf);
	mesg->particle.velocity.z                 = GET_CHAR(buf);
	mesg->particle.color                      = GET_CHAR(buf);
	mesg->particle.count                      = GET_CHAR(buf);
	break;
    case MESG_DAMAGE:
	mesg->damage.type                         = GET_CHAR(buf);
	mesg->damage.armor_damage                 = GET_CHAR(buf);
	mesg->damage.health_damage                = GET_CHAR(buf);
	mesg->damage.position.x                   = GET_COORD(buf);
	mesg->damage.position.y                   = GET_COORD(buf);
	mesg->damage.position.z                   = GET_COORD(buf);
	break;
    case MESG_SPAWN_STATIC:
	mesg->spawn_static.type                   = GET_CHAR(buf);
	mesg->spawn_static.default_model          = GET_CHAR(buf);
	mesg->spawn_static.default_frame          = GET_CHAR(buf);
	mesg->spawn_static.default_colormap       = GET_CHAR(buf);
	mesg->spawn_static.default_skin           = GET_CHAR(buf);
	mesg->spawn_static.default_position.x     = GET_COORD(buf);
	mesg->spawn_static.default_orientation.p  = GET_ANGLE(buf);
	mesg->spawn_static.default_position.y     = GET_COORD(buf);
	mesg->spawn_static.default_orientation.y  = GET_ANGLE(buf);
	mesg->spawn_static.default_position.z     = GET_COORD(buf);
	mesg->spawn_static.default_orientation.r  = GET_ANGLE(buf);
	break;
    case MESG_SPAWN_BINARY:
	mesg->spawn_binary.type                   = GET_CHAR(buf);
	break;
    case MESG_SPAWN_BASELINE:
	mesg->spawn_baseline.type                 = GET_CHAR(buf);
	mesg->spawn_baseline.entity               = GET_SHORT(buf);
	mesg->spawn_baseline.default_model        = GET_CHAR(buf);
	mesg->spawn_baseline.default_frame        = GET_CHAR(buf);
	mesg->spawn_baseline.default_colormap
	                                          = GET_CHAR(buf);
	mesg->spawn_baseline.default_skin         = GET_CHAR(buf);
	mesg->spawn_baseline.default_position.x   = GET_COORD(buf);
	mesg->spawn_baseline.default_orientation.p  
	                                          = GET_ANGLE(buf);
	mesg->spawn_baseline.default_position.y   = GET_COORD(buf);
	mesg->spawn_baseline.default_orientation.y  
	                                          = GET_ANGLE(buf);
	mesg->spawn_baseline.default_position.z   = GET_COORD(buf);
	mesg->spawn_baseline.default_orientation.r  
	                                          = GET_ANGLE(buf);
	break;
    case MESG_TEMP_ENTITY:
	mesg->temp_entity.type                    = GET_CHAR(buf);
	mesg->temp_entity.entity_type             = GET_CHAR(buf);
	switch (mesg->temp_entity.entity_type) {
	case 0: case 1: case 2: case 3: case 4:
	case 7: case 8: case 10: case 11:
	    mesg->temp_entity.position.x          = GET_COORD(buf);
	    mesg->temp_entity.position.y          = GET_COORD(buf);
	    mesg->temp_entity.position.z          = GET_COORD(buf);
	    break;
	case 5: case 6: case 9:
	    mesg->temp_entity.entity              = GET_SHORT(buf);
	    mesg->temp_entity.position.x          = GET_COORD(buf);
	    mesg->temp_entity.position.y          = GET_COORD(buf);
	    mesg->temp_entity.position.z          = GET_COORD(buf);
	    mesg->temp_entity.destination.x       = GET_COORD(buf);
	    mesg->temp_entity.destination.y       = GET_COORD(buf);
	    mesg->temp_entity.destination.z       = GET_COORD(buf);
	    break;
	}
	break;
    case MESG_SET_PAUSE:
	mesg->set_pause.type                      = GET_CHAR(buf);
	mesg->set_pause.pause                     = GET_CHAR(buf);
	break;
    case MESG_SIGNON_NUM:
	mesg->signon_num.type                     = GET_CHAR(buf);
	mesg->signon_num.signon_num               = GET_CHAR(buf);
	break;
    case MESG_CENTER_PRINT:
	mesg->center_print.type                   = GET_CHAR(buf);
	mesg->center_print.text                   = GET_STRING(buf);
	break;
    case MESG_KILLED_MONSTER:
	mesg->killed_monster.type                 = GET_CHAR(buf);
	break;
    case MESG_FOUND_SECRET:
	mesg->found_secret.type                   = GET_CHAR(buf);
	break;
    case MESG_SPAWN_STATIC_SOUND:
	mesg->spawn_static_sound.type             = GET_CHAR(buf);
	mesg->spawn_static_sound.position.x       = GET_COORD(buf);
	mesg->spawn_static_sound.position.y       = GET_COORD(buf);
	mesg->spawn_static_sound.position.z       = GET_COORD(buf);
	mesg->spawn_static_sound.sound            = GET_CHAR(buf);
	mesg->spawn_static_sound.volume           = GET_VOLUME(buf);
	mesg->spawn_static_sound.attenuation      = GET_CHAR(buf) >> 6;
	break;
    case MESG_INTERMISSION:
	mesg->intermission.type                   = GET_CHAR(buf);
	break;
    case MESG_FINALE:
	mesg->finale.type                         = GET_CHAR(buf);
	mesg->finale.text                         = GET_STRING(buf);
	break;
    case MESG_CD_TRACK:
	mesg->cd_track.type                       = GET_CHAR(buf);
	mesg->cd_track.first_track                = GET_CHAR(buf);
	mesg->cd_track.last_track                 = GET_CHAR(buf);
	break;
    case MESG_SELL_SCREEN:
	mesg->sell_screen.type                    = GET_CHAR(buf);
	break;
    default:
	if (*buf & 0x80) {
	    mesg->update_entity.type              = MESG_UPDATE_ENTITY;
	    mesg->update_entity.mask = mask       = GET_CHAR(buf) & 0x7f;
	    mesg->update_entity.mask = mask       = mask | ((mask & 0x0001) ?
					            GET_CHAR(buf) << 8 : 0);
	    mesg->update_entity.entity            = (mask & 0x4000) ?
		                                    GET_SHORT(buf) :
		                                    GET_CHAR(buf);
	    mesg->update_entity.model             = (mask & 0x0400) ?
		                                    GET_CHAR(buf) : 0;
	    mesg->update_entity.frame             = (mask & 0x0040) ?
		                                    GET_CHAR(buf) : 0;
	    mesg->update_entity.colormap          = (mask & 0x0800) ?
		                                    GET_CHAR(buf) : 0;
	    mesg->update_entity.skin              = (mask & 0x1000) ?
		                                    GET_CHAR(buf) : 0;
	    mesg->update_entity.attack_state      = (mask & 0x2000) ?
		                                    GET_CHAR(buf) : 0;
	    mesg->update_entity.position.x        = (mask & 0x0002) ?
		                                    GET_COORD(buf) : 0.0;
	    mesg->update_entity.orientation.p     = (mask & 0x0100) ?
		                                    GET_ANGLE(buf) : 0.0;
	    mesg->update_entity.position.y        = (mask & 0x0004) ?
		                                    GET_COORD(buf) : 0.0;
	    mesg->update_entity.orientation.y     = (mask & 0x0010) ?
		                                    GET_ANGLE(buf) : 0.0;
	    mesg->update_entity.position.z        = (mask & 0x0008) ?
		                                    GET_COORD(buf) : 0.0;
	    mesg->update_entity.orientation.r     = (mask & 0x0200) ?
		                                    GET_ANGLE(buf) : 0.0;
	    mesg->update_entity.new               = ((mask & 0x0020) != 0);
	}
	else {
	    mesg->bad.type                        = MESG_BAD;
	}
	break;
    }

    return buf - start;
}

int
sv_to_cl_mesg_put(char *buf, sv_to_cl_mesg_t *mesg)
{
    int   i;
    short mask;
    char *start = buf;

    switch (mesg->type) {
    case MESG_NOP:
	PUT_CHAR(buf, mesg->nop.type);
	break;
    case MESG_DISCONNECT:
	PUT_CHAR(buf, mesg->disconnect.type);
	break;
    case MESG_UPDATE_STAT:
	PUT_CHAR(buf, mesg->update_stat.type);
	PUT_CHAR(buf, mesg->update_stat.index);
	PUT_LONG(buf, mesg->update_stat.value);
	break;
    case MESG_VERSION:
	PUT_CHAR(buf, mesg->version.type);
	PUT_LONG(buf, mesg->version.version);
	break;
    case MESG_SET_VIEW:
	PUT_CHAR(buf, mesg->set_view.type);
	PUT_SHORT(buf, mesg->set_view.entity);
	break;
    case MESG_SOUND:
	mask = mesg->sound.mask;
	if (mesg->sound.volume      != 1.0f) mask |= 0x0001;
	if (mesg->sound.attenuation !=    1) mask |= 0x0002;
	PUT_CHAR(buf, mesg->sound.type);
	PUT_CHAR(buf, mask);
	if (mask & 0x0001) PUT_VOLUME(buf, mesg->sound.volume);
	if (mask & 0x0002) PUT_CHAR(buf, mesg->sound.attenuation << 6);
	PUT_SHORT(buf, ((mesg->sound.entity  & 0x1fff) << 3) |
		       ((mesg->sound.channel & 0x0007) << 0));
	PUT_CHAR(buf, mesg->sound.sound);
	PUT_COORD(buf, mesg->sound.position.x);
	PUT_COORD(buf, mesg->sound.position.y);
	PUT_COORD(buf, mesg->sound.position.z);
	break ;
    case MESG_TIME:
	PUT_CHAR(buf, mesg->time.type);
	PUT_FLOAT(buf, mesg->time.time);
	break;
    case MESG_PRINT:
	PUT_CHAR(buf, mesg->print.type);
	PUT_STRING(buf, mesg->print.text);
	break;
    case MESG_STUFF_TEXT:
	PUT_CHAR(buf, mesg->stuff_text.type);
	PUT_STRING(buf, mesg->stuff_text.text);
	break;
    case MESG_SET_ANGLE:
	PUT_CHAR(buf, mesg->set_angle.type);
	PUT_ANGLE(buf, mesg->set_angle.orientation.p);
	PUT_ANGLE(buf, mesg->set_angle.orientation.y);
	PUT_ANGLE(buf, mesg->set_angle.orientation.r);
	break;
    case MESG_SERVER_INFO:
	PUT_CHAR(buf, mesg->server_info.type);
	PUT_LONG(buf, mesg->server_info.version);
	PUT_CHAR(buf, mesg->server_info.max_clients);
	PUT_CHAR(buf, mesg->server_info.multi_player);
	PUT_STRING(buf, mesg->server_info.map_name);
	for (i = 1; i <= mesg->server_info.num_models; i++)
	    PUT_STRING(buf, mesg->server_info.model_names[i]);
	PUT_CHAR(buf, 0x00);
	for (i = 1; i <= mesg->server_info.num_sounds; i++)
	    PUT_STRING(buf, mesg->server_info.sound_names[i]);
	PUT_CHAR(buf, 0x00);
	break;
    case MESG_LIGHT_STYLE:
	PUT_CHAR(buf, mesg->light_style.type);
	PUT_CHAR(buf, mesg->light_style.style);
	PUT_STRING(buf, mesg->light_style.text);
	break;
    case MESG_UPDATE_NAME:
	PUT_CHAR(buf, mesg->update_name.type);
	PUT_CHAR(buf, mesg->update_name.player);
	PUT_STRING(buf, mesg->update_name.name);
	break;
    case MESG_UPDATE_FRAGS:
	PUT_CHAR(buf, mesg->update_frags.type);
	PUT_CHAR(buf, mesg->update_frags.player);
	PUT_SHORT(buf, mesg->update_frags.frags);
	break;
    case MESG_CLIENT_DATA:
	mask  = mesg->client_data.mask;
	mask |= (mesg->client_data.bits << 10) & 0x0c00;
	if (mesg->client_data.view_z_offset     !=  22.0f) mask |= 0x0001;
	if (mesg->client_data.view_pitch_offset !=   0.0f) mask |= 0x0002;
	if (mesg->client_data.orientation.p     !=   0.0f) mask |= 0x0004;
	if (mesg->client_data.orientation.y     !=   0.0f) mask |= 0x0008;
	if (mesg->client_data.orientation.r     !=   0.0f) mask |= 0x0010;
	if (mesg->client_data.velocity.x        !=   0.0f) mask |= 0x0020;
	if (mesg->client_data.velocity.y        !=   0.0f) mask |= 0x0040;
	if (mesg->client_data.velocity.z        !=   0.0f) mask |= 0x0080;
	if (mesg->client_data.items             != 0x1001) mask |= 0x0200;
	if (mesg->client_data.weapon_frame      !=      0) mask |= 0x1000;
	if (mesg->client_data.armor             !=      0) mask |= 0x2000;
	if (mesg->client_data.weapon_model      !=      0) mask |= 0x4000;
	PUT_CHAR(buf, mesg->client_data.type);
	PUT_SHORT(buf, mask);
	if (mask & 0x0001) PUT_CHAR(buf, mesg->client_data.view_z_offset);
	if (mask & 0x0002) PUT_CHAR(buf, mesg->client_data.view_pitch_offset);
	if (mask & 0x0004) PUT_ANGLE(buf, mesg->client_data.orientation.p);
	if (mask & 0x0020) PUT_CVELOCITY(buf, mesg->client_data.velocity.x);
	if (mask & 0x0008) PUT_ANGLE(buf, mesg->client_data.orientation.y);
	if (mask & 0x0040) PUT_CVELOCITY(buf, mesg->client_data.velocity.y);
	if (mask & 0x0010) PUT_ANGLE(buf, mesg->client_data.orientation.r);
	if (mask & 0x0080) PUT_CVELOCITY(buf, mesg->client_data.velocity.z);
	if (mask & 0x0200) PUT_LONG(buf, mesg->client_data.items);
	if (mask & 0x1000) PUT_CHAR(buf, mesg->client_data.weapon_frame);
	if (mask & 0x2000) PUT_CHAR(buf, mesg->client_data.armor);
	if (mask & 0x4000) PUT_CHAR(buf, mesg->client_data.weapon_model);
	PUT_SHORT(buf, mesg->client_data.health);
	PUT_CHAR(buf, mesg->client_data.current_ammo);
	PUT_CHAR(buf, mesg->client_data.ammo_shells);
	PUT_CHAR(buf, mesg->client_data.ammo_nails);
	PUT_CHAR(buf, mesg->client_data.ammo_rockets);
	PUT_CHAR(buf, mesg->client_data.ammo_cells);
	PUT_CHAR(buf, mesg->client_data.current_weapon);
	break;
    case MESG_STOP_SOUND:
	PUT_CHAR(buf, mesg->stop_sound.type);
	PUT_SHORT(buf, mesg->stop_sound.sound);
	break;
    case MESG_UPDATE_COLORS:
	PUT_CHAR(buf, mesg->update_colors.type);
	PUT_CHAR(buf, mesg->update_colors.player);
	PUT_CHAR(buf, ((mesg->update_colors.shirt_color & 0x0f) << 4) |
		      ((mesg->update_colors.pants_color & 0x0f) << 0));
	break;
    case MESG_PARTICLE:
	PUT_CHAR(buf, mesg->particle.type);
	PUT_COORD(buf, mesg->particle.position.x);
	PUT_COORD(buf, mesg->particle.position.y);
	PUT_COORD(buf, mesg->particle.position.z);
	PUT_CHAR(buf, mesg->particle.velocity.x);
	PUT_CHAR(buf, mesg->particle.velocity.y);
	PUT_CHAR(buf, mesg->particle.velocity.z);
	PUT_CHAR(buf, mesg->particle.color);
	PUT_CHAR(buf, mesg->particle.count);
	break;
    case MESG_DAMAGE:
	PUT_CHAR(buf, mesg->damage.type);
	PUT_CHAR(buf, mesg->damage.armor_damage);
	PUT_CHAR(buf, mesg->damage.health_damage);
	PUT_COORD(buf, mesg->damage.position.x);
	PUT_COORD(buf, mesg->damage.position.y);
	PUT_COORD(buf, mesg->damage.position.z);
	break;
    case MESG_SPAWN_STATIC:
	PUT_CHAR(buf, mesg->spawn_static.type);
	PUT_CHAR(buf, mesg->spawn_static.default_model);
	PUT_CHAR(buf, mesg->spawn_static.default_frame);
	PUT_CHAR(buf, mesg->spawn_static.default_colormap);
	PUT_CHAR(buf, mesg->spawn_static.default_skin);
	PUT_COORD(buf, mesg->spawn_static.default_position.x);
	PUT_ANGLE(buf, mesg->spawn_static.default_orientation.p);
	PUT_COORD(buf, mesg->spawn_static.default_position.y);
	PUT_ANGLE(buf, mesg->spawn_static.default_orientation.y);
	PUT_COORD(buf, mesg->spawn_static.default_position.z);
	PUT_ANGLE(buf, mesg->spawn_static.default_orientation.r);
	break;
    case MESG_SPAWN_BINARY:
	PUT_CHAR(buf, mesg->spawn_binary.type);
	break;
    case MESG_SPAWN_BASELINE:
	PUT_CHAR(buf, mesg->spawn_baseline.type);
	PUT_SHORT(buf, mesg->spawn_baseline.entity);
	PUT_CHAR(buf, mesg->spawn_baseline.default_model);
	PUT_CHAR(buf, mesg->spawn_baseline.default_frame);
	PUT_CHAR(buf, mesg->spawn_baseline.default_colormap);
	PUT_CHAR(buf, mesg->spawn_baseline.default_skin);
	PUT_COORD(buf, mesg->spawn_baseline.default_position.x);
	PUT_ANGLE(buf, mesg->spawn_baseline.default_orientation.p);
	PUT_COORD(buf, mesg->spawn_baseline.default_position.y);
	PUT_ANGLE(buf, mesg->spawn_baseline.default_orientation.y);
	PUT_COORD(buf, mesg->spawn_baseline.default_position.z);
	PUT_ANGLE(buf, mesg->spawn_baseline.default_orientation.r);
	break;
    case MESG_TEMP_ENTITY:
	PUT_CHAR(buf, mesg->temp_entity.type);
	PUT_CHAR(buf, mesg->temp_entity.entity_type);
	switch (mesg->temp_entity.entity_type) {
	case 0: case 1: case 2: case 3: case 4:
	case 7: case 8: case 10: case 11:
	    PUT_COORD(buf, mesg->temp_entity.position.x);
	    PUT_COORD(buf, mesg->temp_entity.position.y);
	    PUT_COORD(buf, mesg->temp_entity.position.z);
	    break;
	case 5: case 6: case 9:
	    PUT_SHORT(buf, mesg->temp_entity.entity);
	    PUT_COORD(buf, mesg->temp_entity.position.x);
	    PUT_COORD(buf, mesg->temp_entity.position.y);
	    PUT_COORD(buf, mesg->temp_entity.position.z);
	    PUT_COORD(buf, mesg->temp_entity.destination.x);
	    PUT_COORD(buf, mesg->temp_entity.destination.y);
	    PUT_COORD(buf, mesg->temp_entity.destination.z);
	    break;
	}
	break;
    case MESG_SET_PAUSE:
	PUT_CHAR(buf, mesg->set_pause.type);
	PUT_CHAR(buf, mesg->set_pause.pause);
	break;
    case MESG_SIGNON_NUM:
	PUT_CHAR(buf, mesg->signon_num.type);
	PUT_CHAR(buf, mesg->signon_num.signon_num);
	break;
    case MESG_CENTER_PRINT:
	PUT_CHAR(buf, mesg->center_print.type);
	PUT_STRING(buf, mesg->center_print.text);
	break;
    case MESG_KILLED_MONSTER:
	PUT_CHAR(buf, mesg->killed_monster.type);
	break;
    case MESG_FOUND_SECRET:
	PUT_CHAR(buf, mesg->found_secret.type);
	break;
    case MESG_SPAWN_STATIC_SOUND:
	PUT_CHAR(buf, mesg->spawn_static_sound.type);
	PUT_COORD(buf, mesg->spawn_static_sound.position.x);
	PUT_COORD(buf, mesg->spawn_static_sound.position.y);
	PUT_COORD(buf, mesg->spawn_static_sound.position.z);
	PUT_CHAR(buf, mesg->spawn_static_sound.sound);
	PUT_VOLUME(buf, mesg->spawn_static_sound.volume);
	PUT_CHAR(buf, mesg->spawn_static_sound.attenuation << 6);
	break;
    case MESG_INTERMISSION:
	PUT_CHAR(buf, mesg->intermission.type);
	break;
    case MESG_FINALE:
	PUT_CHAR(buf, mesg->finale.type);
	PUT_STRING(buf, mesg->finale.text);
	break;
    case MESG_CD_TRACK:
	PUT_CHAR(buf, mesg->cd_track.type);
	PUT_CHAR(buf, mesg->cd_track.first_track);
	PUT_CHAR(buf, mesg->cd_track.last_track);
	break;
    case MESG_SELL_SCREEN:
	PUT_CHAR(buf, mesg->sell_screen.type);
	break;
    case MESG_UPDATE_ENTITY:
	mask = mesg->update_entity.mask;
	if (mesg->update_entity.entity & 0xff00) mask |= 0x4000;
	PUT_CHAR(buf, (mesg->update_entity.mask & 0x7f) | 0x80);
	if (mask & 0xff01) PUT_CHAR(buf, mesg->update_entity.mask >> 8);
	if (mask & 0x4000) PUT_SHORT(buf, mesg->update_entity.entity);
	else               PUT_CHAR(buf, mesg->update_entity.entity);
	if (mask & 0x0400) PUT_CHAR(buf, mesg->update_entity.model);
	if (mask & 0x0040) PUT_CHAR(buf, mesg->update_entity.frame);
	if (mask & 0x0800) PUT_CHAR(buf, mesg->update_entity.colormap);
	if (mask & 0x1000) PUT_CHAR(buf, mesg->update_entity.skin);
	if (mask & 0x2000) PUT_CHAR(buf, mesg->update_entity.attack_state);
	if (mask & 0x0002) PUT_COORD(buf, mesg->update_entity.position.x);
	if (mask & 0x0100) PUT_ANGLE(buf, mesg->update_entity.orientation.p);
	if (mask & 0x0004) PUT_COORD(buf, mesg->update_entity.position.y);
	if (mask & 0x0010) PUT_ANGLE(buf, mesg->update_entity.orientation.y);
	if (mask & 0x0008) PUT_COORD(buf, mesg->update_entity.position.z);
	if (mask & 0x0200) PUT_ANGLE(buf, mesg->update_entity.orientation.r);
	break;
    default:
	PUT_CHAR(buf, MESG_BAD);
	break;
    }

    return buf - start;
}

char *
sv_to_cl_mesg_name(sv_to_cl_mesg_t *mesg)
{
    switch (mesg->type) {
    default:
    case MESG_BAD:
	return "MESG_BAD";
    case MESG_NOP:
	return "MESG_NOP";
    case MESG_DISCONNECT:
	return "MESG_DISCONNECT";
    case MESG_UPDATE_STAT:
	return "MESG_UPDATE_STAT";
    case MESG_VERSION:
	return "MESG_VERSION";
    case MESG_SET_VIEW:
	return "MESG_SET_VIEW";
    case MESG_SOUND:
	return "MESG_SOUND";
    case MESG_TIME:
	return "MESG_TIME";
    case MESG_PRINT:
	return "MESG_PRINT";
    case MESG_STUFF_TEXT:
	return "MESG_STUFF_TEXT";
    case MESG_SET_ANGLE:
	return "MESG_SET_ANGLE";
    case MESG_SERVER_INFO:
	return "MESG_SERVER_INFO";
    case MESG_LIGHT_STYLE:
	return "MESG_LIGHT_STYLE";
    case MESG_UPDATE_NAME:
	return "MESG_UPDATE_NAME";
    case MESG_UPDATE_FRAGS:
	return "MESG_UPDATE_FRAGS";
    case MESG_CLIENT_DATA:
	return "MESG_CLIENT_DATA";
    case MESG_STOP_SOUND:
	return "MESG_STOP_SOUND";
    case MESG_UPDATE_COLORS:
	return "MESG_UPDATE_COLORS";
    case MESG_PARTICLE:
	return "MESG_PARTICLE";
    case MESG_DAMAGE:
	return "MESG_DAMAGE";
    case MESG_SPAWN_STATIC:
	return "MESG_SPAWN_STATIC";
    case MESG_SPAWN_BINARY:
	return "MESG_SPAWN_BINARY";
    case MESG_SPAWN_BASELINE:
	return "MESG_SPAWN_BASELINE";
    case MESG_TEMP_ENTITY:
	return "MESG_TEMP_ENTITY";
    case MESG_SET_PAUSE:
	return "MESG_SET_PAUSE";
    case MESG_SIGNON_NUM:
	return "MESG_SIGNON_NUM";
    case MESG_CENTER_PRINT:
	return "MESG_CENTER_PRINT";
    case MESG_KILLED_MONSTER:
	return "MESG_KILLED_MONSTER";
    case MESG_FOUND_SECRET:
	return "MESG_FOUND_SECRET";
    case MESG_SPAWN_STATIC_SOUND:
	return "MESG_SPAWN_STATIC_SOUND";
    case MESG_INTERMISSION:
	return "MESG_INTERMISSION";
    case MESG_FINALE:
	return "MESG_FINALE";
    case MESG_CD_TRACK:
	return "MESG_CD_TRACK";
    case MESG_SELL_SCREEN:
	return "MESG_SELL_SCREEN";
    case MESG_UPDATE_ENTITY:
	return "MESG_UPDATE_ENTITY";
    }
}

void
sv_to_cl_mesg_debug(sv_to_cl_mesg_t *mesg, FILE *file)
{
    char buf[16384];
    int  len;

    switch (mesg->type) {
    default:
	len = sv_to_cl_mesg_put(buf, mesg);
	fprintf(file, "[server message: %s, %d bytes]\n",
		sv_to_cl_mesg_name(mesg), len);
	hex_dump(buf, len, "S", file);
	break;
    }
}

char
sv_to_cl_mesg_peek(char *buf)
{
    switch (*buf) {
    default:
    case MESG_BAD:
	return MESG_BAD;
    case MESG_NOP:
    case MESG_DISCONNECT:
    case MESG_UPDATE_STAT:
    case MESG_VERSION:
    case MESG_SET_VIEW:
    case MESG_SOUND:
    case MESG_TIME:
    case MESG_PRINT:
    case MESG_STUFF_TEXT:
    case MESG_SET_ANGLE:
    case MESG_SERVER_INFO:
    case MESG_LIGHT_STYLE:
    case MESG_UPDATE_NAME:
    case MESG_UPDATE_FRAGS:
    case MESG_CLIENT_DATA:
    case MESG_STOP_SOUND:
    case MESG_UPDATE_COLORS:
    case MESG_PARTICLE:
    case MESG_DAMAGE:
    case MESG_SPAWN_STATIC:
    case MESG_SPAWN_BINARY:
    case MESG_SPAWN_BASELINE:
    case MESG_TEMP_ENTITY:
    case MESG_SET_PAUSE:
    case MESG_SIGNON_NUM:
    case MESG_CENTER_PRINT:
    case MESG_KILLED_MONSTER:
    case MESG_FOUND_SECRET:
    case MESG_SPAWN_STATIC_SOUND:
    case MESG_INTERMISSION:
    case MESG_FINALE:
    case MESG_CD_TRACK:
    case MESG_SELL_SCREEN:
    case MESG_UPDATE_ENTITY:
	return *buf;
    }
}

int
sv_to_cl_mesg_skip(char *buf)
{
    short temp;
    short mask;
    char *start = buf;

    switch (*buf) {
    case MESG_NOP:
	SKIP_CHAR(buf);
	break;
    case MESG_DISCONNECT:
	SKIP_CHAR(buf);
	break;
    case MESG_UPDATE_STAT:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_LONG(buf);
	break;
    case MESG_VERSION:
	SKIP_CHAR(buf);
	SKIP_LONG(buf);
	break;
    case MESG_SET_VIEW:
	SKIP_CHAR(buf);
	SKIP_SHORT(buf);
	break;
    case MESG_SOUND:
	SKIP_CHAR(buf);
	mask  = GET_CHAR(buf);
	if (mask & 0x0001) SKIP_VOLUME(buf);
	if (mask & 0x0002) SKIP_CHAR(buf);
	SKIP_SHORT(buf);
	SKIP_CHAR(buf);
	SKIP_COORD(buf);
	SKIP_COORD(buf);
	SKIP_COORD(buf);
	break ;
    case MESG_TIME:
	SKIP_CHAR(buf);
	SKIP_FLOAT(buf);
	break;
    case MESG_PRINT:
	SKIP_CHAR(buf);
	SKIP_STRING(buf);
	break;
    case MESG_STUFF_TEXT:
	SKIP_CHAR(buf);
	SKIP_STRING(buf);
	break;
    case MESG_SET_ANGLE:
	SKIP_CHAR(buf);
	SKIP_ANGLE(buf);
	SKIP_ANGLE(buf);
	SKIP_ANGLE(buf);
	break;
    case MESG_SERVER_INFO:
	SKIP_CHAR(buf);
	SKIP_LONG(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_STRING(buf);
	do {
	    SKIP_STRING(buf);
	} while (*buf);
	SKIP_CHAR(buf);
	do {
	    SKIP_STRING(buf);
	} while (*buf);
	SKIP_CHAR(buf);
	break;
    case MESG_LIGHT_STYLE:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_STRING(buf);
	break;
    case MESG_UPDATE_NAME:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_STRING(buf);
	break;
    case MESG_UPDATE_FRAGS:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_SHORT(buf);
	break;
    case MESG_CLIENT_DATA:
	SKIP_CHAR(buf);
	mask = GET_SHORT(buf);
	if (mask & 0x0001) SKIP_CHAR(buf);
	if (mask & 0x0002) SKIP_CHAR(buf);
	if (mask & 0x0004) SKIP_ANGLE(buf);
	if (mask & 0x0020) SKIP_CVELOCITY(buf);
	if (mask & 0x0008) SKIP_ANGLE(buf);
	if (mask & 0x0040) SKIP_CVELOCITY(buf);
	if (mask & 0x0010) SKIP_ANGLE(buf);
	if (mask & 0x0080) SKIP_CVELOCITY(buf);
	if (mask & 0x0200) SKIP_LONG(buf);
	if (mask & 0x1000) SKIP_CHAR(buf);
	if (mask & 0x2000) SKIP_CHAR(buf);
	if (mask & 0x4000) SKIP_CHAR(buf);
	SKIP_SHORT(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	break;
    case MESG_STOP_SOUND:
	SKIP_CHAR(buf);
	SKIP_SHORT(buf);
	break;
    case MESG_UPDATE_COLORS:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	break;
    case MESG_PARTICLE:
	SKIP_CHAR(buf);
	SKIP_COORD(buf);
	SKIP_COORD(buf);
	SKIP_COORD(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	break;
    case MESG_DAMAGE:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_COORD(buf);
	SKIP_COORD(buf);
	SKIP_COORD(buf);
	break;
    case MESG_SPAWN_STATIC:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_COORD(buf);
	SKIP_ANGLE(buf);
	SKIP_COORD(buf);
	SKIP_ANGLE(buf);
	SKIP_COORD(buf);
	SKIP_ANGLE(buf);
	break;
    case MESG_SPAWN_BINARY:
	SKIP_CHAR(buf);
	break;
    case MESG_SPAWN_BASELINE:
	SKIP_CHAR(buf);
	SKIP_SHORT(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_COORD(buf);
	SKIP_ANGLE(buf);
	SKIP_COORD(buf);
	SKIP_ANGLE(buf);
	SKIP_COORD(buf);
	SKIP_ANGLE(buf);
	break;
    case MESG_TEMP_ENTITY:
	SKIP_CHAR(buf);
	temp = GET_CHAR(buf);
	switch (temp) {
	case 0: case 1: case 2: case 3: case 4:
	case 7: case 8: case 10: case 11:
	    SKIP_COORD(buf);
	    SKIP_COORD(buf);
	    SKIP_COORD(buf);
	    break;
	case 5: case 6: case 9:
	    SKIP_SHORT(buf);
	    SKIP_COORD(buf);
	    SKIP_COORD(buf);
	    SKIP_COORD(buf);
	    SKIP_COORD(buf);
	    SKIP_COORD(buf);
	    SKIP_COORD(buf);
	    break;
	}
	break;
    case MESG_SET_PAUSE:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	break;
    case MESG_SIGNON_NUM:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	break;
    case MESG_CENTER_PRINT:
	SKIP_CHAR(buf);
	SKIP_STRING(buf);
	break;
    case MESG_KILLED_MONSTER:
	SKIP_CHAR(buf);
	break;
    case MESG_FOUND_SECRET:
	SKIP_CHAR(buf);
	break;
    case MESG_SPAWN_STATIC_SOUND:
	SKIP_CHAR(buf);
	SKIP_COORD(buf);
	SKIP_COORD(buf);
	SKIP_COORD(buf);
	SKIP_CHAR(buf);
	SKIP_VOLUME(buf);
	SKIP_CHAR(buf);
	break;
    case MESG_INTERMISSION:
	SKIP_CHAR(buf);
	break;
    case MESG_FINALE:
	SKIP_CHAR(buf);
	SKIP_STRING(buf);
	break;
    case MESG_CD_TRACK:
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	SKIP_CHAR(buf);
	break;
    case MESG_SELL_SCREEN:
	SKIP_CHAR(buf);
	break;
    default:
	if (*buf & 0x80) {
	    mask  = GET_CHAR(buf) & 0x7f;
	    if (mask & 0x0001) mask |= GET_CHAR(buf) << 8;
	    if (mask & 0x4000) SKIP_SHORT(buf);
	    else               SKIP_CHAR(buf);
	    if (mask & 0x0400) SKIP_CHAR(buf);
	    if (mask & 0x0040) SKIP_CHAR(buf);
	    if (mask & 0x0800) SKIP_CHAR(buf);
	    if (mask & 0x1000) SKIP_CHAR(buf);
	    if (mask & 0x2000) SKIP_CHAR(buf);
	    if (mask & 0x0002) SKIP_COORD(buf);
	    if (mask & 0x0100) SKIP_ANGLE(buf);
	    if (mask & 0x0004) SKIP_COORD(buf);
	    if (mask & 0x0010) SKIP_ANGLE(buf);
	    if (mask & 0x0008) SKIP_COORD(buf);
	    if (mask & 0x0200) SKIP_ANGLE(buf);
	}
	else {
	    buf = start;
	}
	break;
    }

    return buf - start;
}
